/*
** VideoImage
**	This module contains functions and utilities to convert the
**	raw video frame (Y, Cb, Cr) into XImage structure.
**
** Copyright 1993, 1994 by Markku Savela and
**	Technical Research Centre of Finland
*/
#include <stdio.h>
#if ANSI_INCLUDES
#	include <stddef.h>
#	include <stdlib.h>
#endif
#include <assert.h>
#include <math.h>

#include <X11/Xlib.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xew/VideoP.h>
#include "BasicColor.h"
#include "ImageTools.h"
#include "VideoImage.h"

#ifdef SH_MEM
#	include <sys/ipc.h>
#	include <sys/shm.h>
#	include <X11/extensions/XShm.h>
#endif

/*
** MirrorRotation
**	Setup line, inc and start parameters according to the rotation
**	and mirror parameters.
*/
static void MirrorRotation(rot, mirror, src, dst)
int rot, mirror;
XeImageChannel *src, *dst;
    {
	int last_col = (src->w - 1) * src->inc;
	int last_row = (src->h - 1) * src->line;

	while (rot < 0)
		rot += 360;
	rot = (((rot % 360 + 45)) / 90) * 90;
	if (rot == 0)
	    {
		dst->w = src->w;
		dst->h = src->h;
		dst->line = src->line;
		dst->inc = mirror ? -src->inc : src->inc;
		dst->addr = &src->addr[(mirror ? last_col : 0)];
	    }
	else if (rot == 90)
	    {
		dst->w = src->h;
		dst->h = src->w;
		dst->line = mirror ? -src->inc : src->inc;
		dst->inc = -src->line;
		dst->addr = &src->addr[last_row + (mirror ? 0 : last_col)];
	    }
	else if (rot == 180)
	    {
		dst->w = src->w;
		dst->h = src->h;
		dst->line = -src->line;
		dst->inc = mirror ? src->inc : -src->inc;
		dst->addr = &src->addr[last_row + (mirror ? last_col : 0)];
	    }
	else if (rot == 270)
	    {
		dst->w = src->h;
		dst->h = src->w;
		dst->line = mirror ? src->inc : -src->inc;
		dst->inc = src->line;
		dst->addr = &src->addr[mirror ? 0 : last_col];
	    }
    }

/*
** XeVideoImage
**	This function is called once per each video frame
*/
void XeVideoImage(w)
XeVideoWidget w;	/* Video Widget */
    {
	XImage *image;
	XeImageChannel data[4];
	XeRawImage *raw = w->video.raw_image;
	int i;

	if (!XtIsRealized((Widget)w) || raw == NULL || !raw->num_channels)
		return;
	if (!w->video.done)
	    {
		Boolean resize = w->basic.resize;

		_XeDestroyImage((XeBasicWidget)w, w->video.dsp_image);
		w->basic.resize = w->video.set_width || w->video.set_height;
		w->video.dsp_image =
			_XeCreateImage((XeBasicWidget)w, raw, w->video.dither,
				       w->video.color_quantize);
		w->video.done = True;
		w->video.set_width = False;
		w->video.set_height = False;
		w->basic.resize = resize;
	    }
	if (!w->video.dsp_image || (image = w->video.dsp_image->image) == NULL)
		return;
	data[0].addr = (unsigned char *)image->data;
	data[0].w = image->width;
	data[0].h = image->height;
	data[0].line = image->bytes_per_line;
	data[0].inc = (image->bits_per_pixel + 7) / 8;
	for (i = 1; i <= raw->num_channels && i < XtNumber(data); ++i)
		MirrorRotation(w->basic.rotation,w->basic.mirror_image,
			       &raw->channel[i-1], &data[i]);
	_XeBuildImage((XeBasicWidget)w, w->video.dsp_image, i, data);
	/*
	** Call the expose for all of the window.
	*/
	(*w->core.widget_class->core_class.expose)
		((Widget)w, (XEvent *)NULL, (Region)0);
	XFlush(XtDisplay(w));
    }

